/************************************************************
  Copyright (C) 2005-2023 RobSmithDev
  https://robsmithdev.co.uk
*************************************************************/  


#include "stdafx.h"
#include "CloudCreator.h"
#include <math.h>

// Noise Generator
int CCloud::Noise1(int x,int y, int seed)
{
    int n = (x * 11)  + (y * 57);
    n = ((n<<13) ^ n) + seed;
    return ((n*n*n) & 0x7fffffff) >> 24;
}

// Build lookup table for cloud calculations
void CCloud::CalcLookups(UCHAR SkyCover,UCHAR bCloudSharpness)
{
	int x,y;
	float CloudSharpness;

	CloudSharpness = (0.90f) + (bCloudSharpness/2550.0f);

	SkyCover=255-SkyCover;
	for (x=0; x<256; x++)
	{
		y = x - SkyCover;
		if (y>255) y=514-y;
		if (y<0) y=0;
		SkyCoverTable[x] = 255-(int)(pow(CloudSharpness,y) * 255.0f);
	}
}

// Tile "src" in this image
bool CCloud::CreateFromTile(int width,int height,CCloud* src)
{
	if (!CreateCloudMemory(width,height)) return false;
	UCHAR* pCloudDest = pCloud;
	UCHAR* pCloudSrc = pCloud;
	UCHAR* pCloudSrcLineStart;
	int x,y;
	int sy,wid,hei;
	wid = src->GetWidth();
	hei = src->GetWidth();

	width/=wid;

	// Copy the cloud image in over and over
	sy=0;
	pCloudSrcLineStart = src->GetCloudPointer();

	for (y=0; y<height; y++)
	{
		sy++;
		if (sy>=hei)
		{
			sy=0;
			pCloudSrcLineStart = src->GetCloudPointer();
		} else pCloudSrcLineStart+=wid;

		for (x=0; x<width; x++)
		{
			memcpy(pCloudDest,pCloudSrcLineStart,wid);
			pCloudDest+=wid;
		}		
	}

	return true;
}

int CCloud::GetWidth()
{
	return thisW;
}

int CCloud::GetHeight()
{
	return thisH;
}


// Create cloud layer
bool CCloud::CreateClouds(int CloudDetail, int Width,int Height,int CloudSeed,UCHAR SkyCover,UCHAR bCloudSharpness)
{
	CCloud obBaseNoise,obBaseNoiseTemp;
	CCloud obTextures[CLOUD_MAX_LAYERS];
	CCloud *c1, *c2, *cSwapTemp;
	UCHAR *pDestCloud;
	UCHAR* pSourceCloud[CLOUD_MAX_LAYERS];
	int Level,x,y,count,darkest,lightest,tmp;

	if (CloudDetail<0) CloudDetail=0;
	if (CloudDetail>=CLOUD_MAX_LAYERS) CloudDetail=CLOUD_MAX_LAYERS-1;

	// Calculate the size of the smallest tile
	int SmallestW	=	Width >> CloudDetail;
	int SmallestH	=	Height >> CloudDetail;
	Width = SmallestW << CloudDetail;
	Height = SmallestH << CloudDetail;

	// Prepare lookup table for sky cover
	CalcLookups(SkyCover,bCloudSharpness);

	// Create noisy picture to use as the base for our cloud
	if (!obBaseNoise.CreateNoise(CloudSeed,SmallestW,SmallestH)) return false;

	// Scale and tile
	c1 = &obBaseNoise;
	c2 = &obBaseNoiseTemp;
	for (Level=0; Level<=CloudDetail; Level++)
	{
		if (Level)  // Level 0 is not scaled up and is tiled from there
		{
			if (!c1->SampleUpCloud(c2)) return false;

			// Swap over the pointers
			cSwapTemp = c1;
			c1 = c2;
			c2 = cSwapTemp;
		}
		// C1 is always the cloud image we want here

		if (!obTextures[Level].CreateFromTile(Width,Height,c1)) return false;
		pSourceCloud[Level] = obTextures[Level].GetCloudPointer();
	}

	// Create the returning cloud buffer
	if (!CreateCloudMemory(Width,Height)) return false;

	// Combine the cloud images whilst working out the darkest and lighest areas
	darkest = 255;
	lightest = 0;
	pDestCloud = pCloud;   // Write into our cloud without messing up the pointer
	tmp = CloudDetail+1;

	for (y=0; y<Height; y++)
		for (x=0; x<Width; x++)
		{
			count = 0;
			for (Level=0; Level<=CloudDetail; Level++)
			{
				count+=(*pSourceCloud[Level]) >> (tmp-Level);
				pSourceCloud[Level]++;
			}

			// Maintin the lightest and darkest values
			if (count<darkest) darkest=count;
			if (count>lightest) lightest=count;

			// Write out the final result
			*pDestCloud = count;  
			pDestCloud++;
		}

	// Now correct the image to give the biggest brightness range
	pDestCloud = pCloud;   // Write into our cloud without messing up the pointer

	if (lightest==darkest) lightest=32768; else lightest = 65535/(lightest-darkest);
	

	for (y=0; y<Height; y++)
		for (x=0; x<Width; x++)
		{
			*pDestCloud = SkyCoverTable[((*pDestCloud-darkest)*lightest) >> 8];
			pDestCloud++;
		}
	
	return true;

}

CCloud::CCloud()
{
	pCloud=NULL;
	thisW=0;
	thisH=0;
}

CCloud::~CCloud()
{
	FreeCloud();
}

UCHAR* CCloud::GetCloudPointer()
{
	return pCloud;
}

bool CCloud::CreateCloudMemory(int width,int height)
{
	if ((pCloud)&&(thisW==width)&&(thisH==height)) return true;   // dont need to re-allocate

	FreeCloud();
	thisW = width;
	thisH = height;

	pCloud = (UCHAR*) malloc(width*height);

	return pCloud!=NULL;
}

// Convert the image to a bitmap
HBITMAP CCloud::ConvertToBitmap(DWORD BackgroundColor)
{
	BITMAPINFO BitmapInfo;

	int x,y,rev,precalc;
	UCHAR* pCloudSrc = pCloud;
	RGBQUAD* pBitmapBits;
	int bgR = GetRValue(BackgroundColor);
	int bgG = GetGValue(BackgroundColor);
	int bgB = GetBValue(BackgroundColor);
	HBITMAP hBitmap;

	if (pCloudSrc==NULL) return 0;

	BitmapInfo.bmiHeader.biSize = sizeof(BitmapInfo.bmiHeader);
	BitmapInfo.bmiHeader.biWidth = thisW;
	BitmapInfo.bmiHeader.biHeight = thisH;
	BitmapInfo.bmiHeader.biPlanes = 1;
	BitmapInfo.bmiHeader.biBitCount = 32;
	BitmapInfo.bmiHeader.biCompression = BI_RGB;

	hBitmap = CreateDIBSection(0,&BitmapInfo,DIB_RGB_COLORS,(void**)&pBitmapBits,NULL,0);
	if (hBitmap==0) return hBitmap;
	
	// Convert to bitmap by mixing with white and the background colour
	for (y=0; y<thisH; y++)
		for (x=0; x<thisW; x++)
		{
			rev = 255-*pCloudSrc;
			precalc = (255**pCloudSrc);

			pBitmapBits->rgbBlue =  ((bgB*rev) + precalc) >> 8;
			pBitmapBits->rgbGreen = ((bgG*rev) + precalc) >> 8;
			pBitmapBits->rgbRed =   ((bgR*rev) + precalc) >> 8;

			pCloudSrc++;
			pBitmapBits++;
		}

	return hBitmap;
}

// Joins clouds together
HBITMAP CCloud::AnimateClouds(DWORD SkyColor,UCHAR CloudCover, CCloud* pCloudList, ...)
{
	// Merge all cloudies together
	CCloud* Clouds[CLOUD_MAX_LAYERS];
	int nCloudCount = 1;
	Clouds[0] = pCloudList;
	va_list marker;
	CCloud* pTemp;
	BITMAPINFO BitmapInfo;
	RGBQUAD* pBitmapBits;
	HBITMAP hBitmap;
	int bgR = GetRValue(SkyColor);
	int bgG = GetGValue(SkyColor);
	int bgB = GetBValue(SkyColor);
	int DestR,DestG,DestB;
	UCHAR* pMem[CLOUD_MAX_LAYERS];
	int a,x,y,rev,precalc,Alpha,PicCol;

	pMem[0] = pCloudList->GetCloudPointer();


	// Build an array of clouds
	va_start( marker, pCloudList );
	while ((pTemp=va_arg(marker,CCloud*))&&(nCloudCount<CLOUD_MAX_LAYERS))
	{
		Clouds[nCloudCount] = pTemp;
		pMem[nCloudCount] = pTemp->GetCloudPointer();
		nCloudCount++;

		// Ensure they are all the same size
		if ((pTemp->GetWidth()!=pCloudList->GetWidth())||(pTemp->GetHeight()!=pCloudList->GetHeight())) return 0;
	}
	va_end(marker);

	// Now create a bitmap of the size of all of the input clouds
	BitmapInfo.bmiHeader.biSize = sizeof(BitmapInfo.bmiHeader);
	BitmapInfo.bmiHeader.biWidth = pCloudList->GetWidth();
	BitmapInfo.bmiHeader.biHeight = pCloudList->GetHeight();
	BitmapInfo.bmiHeader.biPlanes = 1;
	BitmapInfo.bmiHeader.biBitCount = 32;
	BitmapInfo.bmiHeader.biCompression = BI_RGB;

	hBitmap = CreateDIBSection(0,&BitmapInfo,DIB_RGB_COLORS,(void**)&pBitmapBits,NULL,0);
	if (hBitmap==0) return hBitmap;

	// Combine the colours together to form the cloud bitmap
	for (y=0; y<pCloudList->GetHeight(); y++)
		for (x=0; x<pCloudList->GetWidth(); x++)
		{
			if (CloudCover>128)
			{
				Alpha = (CloudCover-128)<<2;
				if (Alpha>255) Alpha=255;
				rev = 255-Alpha;


				DestR = ((bgR*rev)+(255*Alpha)) >> 8;
				DestG = ((bgG*rev)+(255*Alpha)) >> 8;
				DestB = ((bgB*rev)+(255*Alpha)) >> 8;

			} else
			{
				DestR = bgR;
				DestG = bgG;
				DestB = bgB;
			}


			
			// Now mix with the pixel colours from the other layers
			for (a=0; a<nCloudCount; a++)
			{
				rev = 255-*pMem[a];
				precalc = (255**pMem[a]);  // cloud colour is white

				DestB =  ((DestB * rev) + precalc) >> 8;
				DestG =  ((DestG * rev) + precalc) >> 8;
				DestR =  ((DestR * rev) + precalc) >> 8;

				if (a<nCloudCount-1) pMem[a]++;
			}

			// Top level cloud is a little different, as when cloudcover is 100% it should only show through
			a=nCloudCount-1;

			if (CloudCover>128)
			{
				Alpha = (CloudCover-128)<<1;
				rev = 255-Alpha;
				PicCol = *pMem[a]*Alpha;   // Colour of dest pixel

				DestB = ((DestB*rev) + (PicCol)) >> 8;
				DestG = ((DestG*rev) + (PicCol)) >> 8;
				DestR = ((DestR*rev) + (PicCol)) >> 8;
			}

			pMem[a]++;

			pBitmapBits->rgbBlue =  DestB;
			pBitmapBits->rgbGreen = DestG;
			pBitmapBits->rgbRed =   DestR;

			pBitmapBits++;
		}

	return hBitmap;
}


// Rolls the cloud information so many pixels in each direction
bool CCloud::RollCloud(int x,int y)
{
	UCHAR* pBuff;
	int yp;
	// Wrap them around the width/heights
	x = x % thisW;
	y = y % thisH;
	if (x<0) x+=thisW;
	if (y<0) y+=thisH;
	UCHAR* pCloudData;

	if (pCloud==NULL) return false;

	if (x>0)
	{
		pBuff = new UCHAR[thisW];  // reserve some temporary memory
		if (pBuff==NULL) return false;

		pCloudData = pCloud;

		// Wrap Horizontally First
		for (yp=0; yp<thisH; yp++)
		{
			memcpy(pBuff,pCloudData,thisW);  // Get a copy of the line

			// copy the bits back after moving them
			memcpy(pCloudData+x,pBuff,thisW-x);
			memcpy(pCloudData,pBuff+thisW-x,x);
			pCloudData+=thisW;
		}
		delete [] pBuff;  // Free up the buffer
	}

	if (y>0)
	{
		pBuff = new UCHAR[thisW*thisH];  // reserve some more temporary memory
		if (pBuff==NULL) return false;

		memcpy(pBuff,pCloud,thisW*thisH);  // get a copy of the memory

		// and move it back, but row-shifted
		y*=thisH;
		memcpy(pCloud,pBuff+y,(thisW*thisH)-y);
		memcpy(pCloud+((thisW*thisH)-y),pBuff,y);

		delete [] pBuff;  // Free up the buffer
	}

	

	return true;
}


bool CCloud::CreateNoise(int time, int width,int height)
{
	int x,y;
	UCHAR * pCloudData;

	if (!CreateCloudMemory(width,height)) return false;

	pCloudData = pCloud;

	for (y=0; y<thisH; y++)
		for (x=0; x<thisW; x++)
		{
			*pCloudData = Noise1(x,y,time);
			pCloudData++;
		}	

	return SmoothImage();
}
		
bool CCloud::SmoothImage()
{
	if (pCloud==NULL) return false;  

	int x,y;
	int yStarts[3];  // offsets for these positions, programmed properly for wrapping
	int xStarts[3];  

	UCHAR* pCloudDest = (UCHAR*) malloc(thisW*thisH);
	UCHAR* pCloudDestTemp = pCloudDest;

	if (pCloudDest==NULL) return false;

	// Smooth image
	for (y=0; y<thisH; y++)
	{
		// Prepare vertical offsets
		yStarts[1] = (y*thisW);
		if (y==0) yStarts[0]=thisW*(thisH-1); else yStarts[0]=yStarts[1]-thisW;
		if (y==thisH-1) yStarts[2]=0; else yStarts[2]=yStarts[1]+thisW;

		for (x=0; x<thisW; x++)
		{
			// Prepare horizontal offsets
			xStarts[1] = x;
			if (x==0) xStarts[0]=thisW-1; else xStarts[0]=x-1;
			if (x==thisW-1) xStarts[2]=0; else xStarts[2]=x+1;

			*pCloudDest = 
				    // Corners
			     	((*(UCHAR*)(pCloud+xStarts[0]+yStarts[0])+*(UCHAR*)(pCloud+xStarts[2]+yStarts[0])+*(UCHAR*)(pCloud+xStarts[0]+yStarts[2])+*(UCHAR*)(pCloud+xStarts[2]+yStarts[2])) >> 4) +
					// Sides
					((*(UCHAR*)(pCloud+xStarts[0]+yStarts[1])+*(UCHAR*)(pCloud+xStarts[1]+yStarts[0])+*(UCHAR*)(pCloud+xStarts[2]+yStarts[1])+*(UCHAR*)(pCloud+xStarts[1]+yStarts[2])) >> 3) +
					// Center
					(*(UCHAR*)(pCloud+xStarts[1]+yStarts[1])>>2);

			pCloudDest++;
		}	
	}

	// Free the old cloud image and store the new one (overwrite pointer)
	free(pCloud);
	pCloud = pCloudDestTemp;

	return true;
}

void CCloud::FreeCloud()
{
	if (pCloud) free(pCloud);
	pCloud = NULL;
}

// Sample UP
bool CCloud::SampleUpCloud(CCloud* dest)
{
	int destW = thisW<<1;
	int destH = thisH<<1;
	if (!dest->CreateCloudMemory(destW,destH)) return false;
	if (pCloud==NULL) return false;

	int x,y,xx,yy;
	int yStarts[3];  // offsets for these positions, programmed properly for wrapping
	int xStarts[3];  

	UCHAR* pCloudDest = dest->GetCloudPointer();

	for (yy=0; yy<destH; yy++)
	{
		y = yy >> 1;
		// Prepare vertical offsets
		yStarts[1] = (y*thisW);
		if (y==0) yStarts[0]=thisW*(thisH-1); else yStarts[0]=yStarts[1]-thisW;
		if (y==thisH-1) yStarts[2]=0; else yStarts[2]=yStarts[1]+thisW;

		for (xx=0; xx<destW; xx++)
		{
			x = xx >> 1;

			// Prepare horizontal offsets
			xStarts[1] = x;
			if (x==0) xStarts[0]=thisW-1; else xStarts[0]=x-1;
			if (x==thisW-1) xStarts[2]=0; else xStarts[2]=x+1;

			*pCloudDest = 
				    // Corners
			     	((*(UCHAR*)(pCloud+xStarts[0]+yStarts[0])+*(UCHAR*)(pCloud+xStarts[2]+yStarts[0])+*(UCHAR*)(pCloud+xStarts[0]+yStarts[2])+*(UCHAR*)(pCloud+xStarts[2]+yStarts[2])) >> 4) +
					// Sides
					((*(UCHAR*)(pCloud+xStarts[0]+yStarts[1])+*(UCHAR*)(pCloud+xStarts[1]+yStarts[0])+*(UCHAR*)(pCloud+xStarts[2]+yStarts[1])+*(UCHAR*)(pCloud+xStarts[1]+yStarts[2])) >> 3) +
					// Center
					(*(UCHAR*)(pCloud+xStarts[1]+yStarts[1])>>2);

			pCloudDest++;
		}	
	}

	return true;	
}
